


// st20cc -c2 -cpp -e manage_keys -rcu getput_key.c

/*
func_277
Refernce 600F4FC0

ROM:600F4FC0                 dd 14Bh
ROM:600F4FC4                 dd sub_6006DB16         ; int sub_6006DB16(arg_01);
ROM:600F4FC8                 db 1, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

New pointer 60035800
Arguments: 
04 - argc
0F -- void*
0F -- void*
03 -- int
0D -- int*

*/

#ifndef WIN32
#define memcpy(target,source,len) __asm {ldabc len,target,source;move;}
typedef (fn)();
#define memmove ((fn*)0x6005CC84)
#else
#include <string.h>
#endif

#include "getput_key.h"
#include "bufhead.h"

// manage data inside emu buffers

int manage_keys(byte* request, byte* bufemu, int mode, int* offset) {
	// request struct :
	// 00 - CA system
	// 01 - index
	// 02 - type/key_id
	// 03 - length
	// 04...len+4 payload
	int len, end_emu;
	end_emu = bufemu[BUF_END+1] + (bufemu[BUF_END]<<8);
	switch(mode) {
		case MODE_GETKEY:
			*offset=get_key(request,bufemu);
			break;
		case MODE_PUTKEY:
			len = request[LEN_BYTE];
			*offset=get_key(request,bufemu);
			if((*offset) < end_emu) { // key found
				if(len != bufemu[*offset+LEN_BYTE]) { // unequal key record
					*offset=del_key(bufemu,*offset);
					end_emu= *offset+len+HEADER_LEN;
				}
			}
			else  {// insert new key
				end_emu= *offset+len+HEADER_LEN;
			}
			if(end_emu > DATA_END) {
				*offset=0x1000; // insufficient space
				break;
			}
			memcpy(*offset+bufemu,request, len+HEADER_LEN);
			bufemu[BUF_END] = end_emu >> 8;
			bufemu[BUF_END+1] = end_emu;
			break;
		case MODE_DELKEY:
			*offset=get_key(request,bufemu);
			if((*offset) < end_emu) { // key found
				*offset=del_key(bufemu,*offset);
			}
			break;
		case MODE_GETPROV:
			*offset=get_provi(request,bufemu);
			break;
		case MODE_PUTPROV:
			len = request[LEN_BYTE];
			*offset=get_provi(request,bufemu);
			if((*offset) < end_emu) { // provi found
				if(len != bufemu[*offset+LEN_BYTE]) { // change soid
					*offset=del_key(bufemu,(*offset)); // do not delete keys for that provi
					end_emu= *offset+len+HEADER_LEN;
				}
			}
			else  {// insert new SOID
			   	end_emu= *offset+len+HEADER_LEN;
			}
			if(end_emu > DATA_END) {
				*offset=0x1000; // insufficient space
				break;
			}
			
			memcpy(*offset+bufemu,request, len+HEADER_LEN);
			bufemu[BUF_END] = end_emu >> 8;
			bufemu[BUF_END+1] = end_emu;
			break;
		case MODE_DELPROV:
			*offset=get_provi(request,bufemu);
			if((*offset) < end_emu) { // provi found
				*offset=del_provi(bufemu,*offset);
			}
			break;
		case MODE_GETNEXT:
			*offset=get_next(request,bufemu,*offset);
			break;
		case MODE_DELALL:
			*offset=del_provi_keys(request,bufemu);
			break;
		default: *offset=0x1000; // unsupported command
	}
	return 0;
}

int get_key(byte* request, byte* bufemu) 
{ // returns: offset of existing record if found, else end_emu
	int i=0,end_emu = bufemu[BUF_END+1] + (bufemu[BUF_END]<<8);
	int casys=request[0], index=request[1], type=request[2];
	
	while(i < end_emu) {
		if((casys == bufemu[i])&&
		   (index == bufemu[i+1])&&
		   (type ==  bufemu[i+2])) break;
		i += (bufemu[i+LEN_BYTE] + HEADER_LEN);
	}
	return i;
}

int get_provi(byte* request, byte* bufemu) 
{ 	// returns: offset of existing record if found, else end_emu
	// index and type bytes of request[] are affected
	int i,j,end_emu = bufemu[BUF_END+1] + (bufemu[BUF_END]<<8);
	int casys=request[0], len=request[LEN_BYTE], found=0;
	byte* soid1 = request+HEADER_LEN, *soid2;
	byte id[256];
	for(i=0;i<256;i++) id[i]=0;
	i=0;
	while(i < end_emu) {
		if((casys == bufemu[i])&&
		   (TYPE_SOID ==  bufemu[i+2])) {
		   	id[bufemu[i+1]] = 1;
		   	for(soid2=bufemu+i+HEADER_LEN, j=0; j<len; j++) {
		   		if(soid1[j]!=soid2[j]) break;
		   	}
		   	if(j==len) {
		   		found = 1;
		   		break;
		   	}
		}
		i += (bufemu[i+LEN_BYTE] + HEADER_LEN);
	}
	// update index value
	if(found) request[1] = bufemu[i+1]; 
	else {
		for(j=0;j<256;j++) {
			// locate free index value
			if(!id[j]) break;
		}
		request[1] = j;
	}
	request[2]=TYPE_SOID; 
	return i;
}

int get_next(byte* request,byte* bufemu,int offset)
{ // returns: offset of next record of requested type, else end_emu
	int i=offset,end_emu = bufemu[BUF_END+1] + (bufemu[BUF_END]<<8);
	int casys=request[0], index=request[1], len=request[LEN_BYTE], type=request[2];
	
	if(len == 1) { // enumerate SOIDs for given CA ID
		while(i < end_emu) {
			if((casys == bufemu[i])) break;
			i += (bufemu[i+LEN_BYTE] + HEADER_LEN);
		}
	}
	else { // enumerate keys for given CA ID and Provi index
		while(i < end_emu) {
			if((casys == bufemu[i])&&
			   (index == bufemu[i+1])&&
			   (bufemu[i+2] !=  TYPE_SOID)) break;
			i += (bufemu[i+LEN_BYTE] + HEADER_LEN);
		}
	}
	return i;
}

int del_key(byte* bufemu, int offset) 
{ // delete single record
	
	int end_emu = bufemu[BUF_END+1] + (bufemu[BUF_END]<<8);
	int len=bufemu[offset+LEN_BYTE]+HEADER_LEN;
	
	memmove(bufemu+offset, bufemu+offset+len, end_emu - offset - len);
	end_emu -= len;
	bufemu[BUF_END] = end_emu >> 8;
	bufemu[BUF_END+1] = end_emu;
	return end_emu;
}

int del_provi(byte* bufemu, int offset)
{ // remove all provider's records from service buffer
	int end_emu;
	byte request[HEADER_LEN];
	request[0]=bufemu[offset];
	request[1]=bufemu[offset+1];
	request[LEN_BYTE]=2; // enumerate keys of provi
	// 1st del SOID record
	del_key(bufemu,offset);
	// then remove all keys of that provi
	return del_provi_keys(request, bufemu);
}

int del_provi_keys(byte* request, byte* bufemu) 
{ // remove all provider's keys from current buffer
	int end_emu = bufemu[BUF_END+1] + (bufemu[BUF_END]<<8), offset;
	request[LEN_BYTE]=2; // enumerate keys of provi
	offset=0;
	while(offset < end_emu) {
		offset=get_next(request,bufemu,offset);
		if(offset < end_emu)
			end_emu = del_key(bufemu, offset);
	}
	return end_emu;
}

