#include "mhw.h"
#include <string.h>
#include <stdio.h>

#ifndef byte
   #define byte unsigned char
#endif

#define peeki(buf, offset) (buf[offset]|(buf[offset+1] << 8)|(buf[offset+2] << 16)|(buf[offset+3] << 24))
#define peeks(buf, offset) (buf[offset]|(buf[offset+1] << 8))

#define SECTOR_HEADER_SIZE 0x20
#define FOLDER_OFFSET   0
#define FOLD_NAME_LEN   10
#define FOLDER_HEADER_SIZE  (FOLD_NAME_LEN+sizeof(short))
#define FOLD_LEN        (FOLD_NAME_LEN)
#define NUM_FILES       (FOLD_LEN+sizeof(short))
#define FILE_OFFSET     (NUM_FILES+sizeof(long))
#define FILE_NAME_LEN   12
#define FILE_PTR        (FILE_OFFSET+FILE_NAME_LEN)
#define FILE_LEN        (FILE_PTR+4)
#define FILE_HEADER_LEN  (FILE_NAME_LEN+(3*sizeof(long))) // file name + 4 bytes of file id, 4 bytes of file offset, 4 bytes of file len
#define FILE_DATA (FOLDER_HEADER_SIZE+FILE_HEADER_LEN)

int read_section(byte *buf, mhw_folder *fold, int sect) {
   short off[16];
   int i, fold_num;
   for(i=0; i < 16; i++){
      off[i] = peeks(buf, i*2);
      if(!off[i]) break;
   }
   fold_num = i;
   for(i = 0; i < fold_num; i++) {
      fold->offset = off[i];
      memcpy(fold->name, buf+fold->offset,FOLD_NAME_LEN);
      fold->len = peeks(buf, fold->offset+FOLD_LEN);
      fold->num_files = peeki(buf, fold->offset+NUM_FILES);
      fold->sect_id = (short)sect;
      fold++;
   }
   return fold_num;
}

//extern mhw_folder *folder_list;
int read_folders(FILE *hnd, mhw_folder *folder) {
    unsigned int nLen,nFolders,nSize,dwSize,nBufAdr,ctrl, i, j, k;
    byte buf[0x10000];
    fseek(hnd,0,2);
    nLen=ftell(hnd);
    fseek(hnd,0,0);
    nSize = nLen/0x10000; // number of sections
    nFolders = 0;
    for(i = 0; i<nSize; i++) {
       //fseek(hnd, i*0x10000, 0);
       fread(buf, 0x10000, 1, hnd);
       j = read_section(buf,&folder[nFolders], i);
       nFolders += j;

    }
    return nFolders;
}

void read_mhwfile(FILE *hnd, mhw_folder *fold, mhw_file *f){
   byte buf[20];
   fseek(hnd, (int)fold->sect_id * 0x10000 + (int)fold->offset + 10 + 2 + 4, 0); // section offset + folder offset + folder name + folder len
   fread(buf,20,1,hnd);
   f->offset = peeki(buf, 12);
   f->len = peeki(buf, 16);
   memcpy(f->name, buf, 12);

}
/*
Offset      0  1  2  3  4  5  6  7   8  9  A  B  C  D  E  F

Sector header: offsets of folders
00020000   20 00 -- 1st folder
		 F4 01 -- 2nd folder
		       00 00  -- last
			     00 00  00 00 00 00 00 00 00 00    ..............
00020010   00 00 00 00 00 00 00 00  00 00 00 00 00 00 00 00   ................

1st folder
00020020   54 53 52 00 00 00 00 00  00 00 -- folder name (10)
					  C8 01 -- folder size 0x1c8
     1st file
						01 00 00 00 -- file_id 1
							      TSR............
00020030   54 53 52 00 00 00 00 00  00 00 00 00 -- file name (12)
						18 00 00 00 -- file data offset
							      TSR.............
00020040   B0 01 00 00 -- file len 0x1b0
		       00 00 00 6B  00 00 55 F0 00 00 6A 3B   ......k..U..j;
00020050   00 00 6B 6C 00 00 74 CC  00 00 75 30 00 00 6D 60   ..kl..t..u0..m`
..................................................
Offset      0  1  2  3  4  5  6  7   8  9  A  B  C  D  E  F

000201F0   00 00 0A F0  2nd folder (offset 0x1f4)
		       45 4D 55 00  00 00 00 00 00 00 -- folder name
						      18 10 -- folder size 0x1018
							      ...EMU.........
		       1st file
00020200   01 00 00 00 -- file_id 1
		       45 4D 55 00  00 00 00 00 00 00 00 00   ....EMU.........
00020210   18 00 00 00 data offset 0x18
		       00 10 00 00  -- data len 0x1000
				    00 6A 00 64 00 12 00 1C   .........j.d....
00020220   00 2A 00 21 00 25 00 65  00 00 00 00 00 00 00 00   .*.!.%.e........
*/

#define pokes(buf, offset, value) {     buf[offset+1] = (value >> 8)& 0xFF; \
                                        buf[offset+0] = value & 0xFF; }
#define pokei(buf, offset, value) {     buf[offset+3] = (value >> 24)& 0xFF; \
                                        buf[offset+2] = (value >> 16)& 0xFF; \
                                        buf[offset+1] = (value >> 8)& 0xFF; \
					buf[offset+0] = value & 0xFF; }

int write_mhw_file(byte* buf, char* fold_name, char* file_name, byte* data, int len) {
   short off[16];
   int i, fold_num, folder_end, free_room;
   // count number of folders
   for(i=0; i < 16; i++){
      off[i] = peeks(buf, i*2);
      if(!off[i]) break;
   }
   fold_num = i;
   if(i) {

	// get next folder offset
	folder_end =  SECTOR_HEADER_SIZE+FOLDER_HEADER_SIZE+peeks(buf, off[i-1]+FOLD_NAME_LEN);
   }
   else {
	folder_end =  SECTOR_HEADER_SIZE;
   }
   free_room = FLASH_SECTOR_SIZE - folder_end;
   if(free_room < (len+FOLDER_HEADER_SIZE+FILE_HEADER_LEN)) return -1;
   // copy folder name and len
   strcpy((char*)(buf + folder_end),fold_name);
   pokes(buf,folder_end+FOLD_LEN, (len+FILE_HEADER_LEN));
   // copy file attributes
   pokei(buf,folder_end+FOLDER_HEADER_SIZE,1); // file id, only one file per folder supported
   strcpy((char*)(buf + folder_end + FILE_OFFSET),file_name);
   pokei(buf, folder_end + FILE_PTR, FILE_HEADER_LEN); // file data offset
   pokei(buf, folder_end + FILE_LEN, len); // file length
   memcpy(buf+folder_end+FOLDER_HEADER_SIZE+FILE_HEADER_LEN,data,len);
   // update sector header
   pokes(buf,fold_num*2,folder_end);
   return 0;
}






 